﻿//
// Copyright (C) 2010 Chris Dziemborowicz
//
// This file is part of Orzeszek Timer.
//
// Orzeszek Timer is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// Orzeszek Timer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Media;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Interop;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Threading;

namespace OrzeszekTimer
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
	{
		private Thread updaterThread;
		private WindowInteropHelper interopHelper;

		private bool notified = true;
		private DateTime start = DateTime.MinValue;
		private DateTime end = DateTime.MinValue;

		public MainWindow()
		{
			InitializeComponent();

			try
			{
				foreach (string soundFilePath in Directory.GetFiles("Sounds", "*.wav"))
				{
					MenuItem soundMenuItem = new MenuItem();
					soundMenuItem.Header = System.IO.Path.GetFileNameWithoutExtension(soundFilePath);
					soundMenuItem.Tag = System.IO.Path.GetFileName(soundFilePath);
					soundMenuItem.IsCheckable = true;
					soundMenuItem.Click += new RoutedEventHandler(SoundMenuItem_Click);

					MainContextMenu.Items.Insert(MainContextMenu.Items.Count - 3, soundMenuItem);
				}
			}
			catch (Exception)
			{
			}

			UpdateSoundMenuItems();

			LoopTimerMenuItem.IsChecked = Settings.Default.LoopTimer;
			RememberTimerOnCloseMenuItem.IsChecked = Settings.Default.RememberTimerOnClose;

			if (Settings.Default.WindowSettingsSaved)
				try
				{
					Height = Settings.Default.WindowHeight;
					Width = Settings.Default.WindowWidth;
					Left = Settings.Default.WindowLeft;
					Top = Settings.Default.WindowTop;
					WindowState = Settings.Default.WindowState;
				}
				catch (Exception)
				{
				}

			interopHelper = new WindowInteropHelper(this);
		}

		private void Window_Loaded(object sender, RoutedEventArgs e)
		{
			string args = (string)Application.Current.Properties["Args"];
			if (args != null)
				try
				{
					object o = FromString(args);

					if (o is DateTime)
					{
						start = DateTime.Now;
						end = (DateTime)o;
						notified = false;

						Settings.Default.LastTimeSpan = TimeSpan.Zero;

						MainTextBox.Template = (ControlTemplate)Resources["ValidTextBoxTemplate"];
						MainButton.Focus();
					}
					else if (o is TimeSpan)
					{
						TimeSpan ts = (TimeSpan)o;

						start = DateTime.Now;
						end = start.Add(ts);
						notified = false;

						Settings.Default.LastTimeSpan = ts;

						MainTextBox.Template = (ControlTemplate)Resources["ValidTextBoxTemplate"];
						MainButton.Focus();
					}

					StartTimer();
				}
				catch (Exception)
				{
					MainTextBox.Template = (ControlTemplate)Resources["InvalidTextBoxTemplate"];
				}
			else if (Settings.Default.RememberTimerOnClose && Settings.Default.TimerRunning && Settings.Default.CurrentStart != DateTime.MinValue && Settings.Default.CurrentEnd != DateTime.MinValue && DateTime.Now > Settings.Default.CurrentStart && DateTime.Now < Settings.Default.CurrentEnd)
			{
				start = Settings.Default.CurrentStart;
				end = Settings.Default.CurrentEnd;
				notified = false;

				MainTextBox.Template = (ControlTemplate)Resources["ValidTextBoxTemplate"];
				MainButton.Focus();

				StartTimer();
			}
			else
			{
				MainTextBox.Text = ToString(Settings.Default.LastTimeSpan.Ticks < 0 ? TimeSpan.Zero : Settings.Default.LastTimeSpan);
				MainTextBox.Focus();
			}
		}

		private void MainTextBox_KeyUp(object sender, KeyEventArgs e)
		{
			if (e.Key == Key.Enter)
				try
				{
					object o = FromString(MainTextBox.Text);

					if (o is DateTime)
					{
						start = DateTime.Now;
						end = (DateTime)o;
						notified = false;

						Settings.Default.LastTimeSpan = TimeSpan.Zero;

						MainTextBox.Template = (ControlTemplate)Resources["ValidTextBoxTemplate"];
						MainButton.Focus();
					}
					else if (o is TimeSpan)
					{
						TimeSpan ts = (TimeSpan)o;

						start = DateTime.Now;
						end = start.Add(ts);
						notified = false;

						Settings.Default.LastTimeSpan = ts;

						MainTextBox.Template = (ControlTemplate)Resources["ValidTextBoxTemplate"];
						MainButton.Focus();
					}
				}
				catch (Exception)
				{
					MainTextBox.Template = (ControlTemplate)Resources["InvalidTextBoxTemplate"];
				}
			else if (e.Key == Key.Escape)
			{
				MainTextBox.Template = (ControlTemplate)Resources["ValidTextBoxTemplate"];
				MainButton.Focus();
			}
		}

		private void MainTextBox_GotFocus(object sender, RoutedEventArgs e)
		{
			MainTextBox.SelectAll();
		}

		private void MainTextBox_LostFocus(object sender, RoutedEventArgs e)
		{
			StartTimer();
		}

		private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
		{
			try
			{
				Settings.Default.LoopTimer = LoopTimerMenuItem.IsChecked;
				Settings.Default.RememberTimerOnClose = RememberTimerOnCloseMenuItem.IsChecked;

				if (Settings.Default.RememberTimerOnClose && start != DateTime.MinValue && end != DateTime.MinValue && DateTime.Now < end)
				{
					Settings.Default.TimerRunning = true;
					Settings.Default.CurrentStart = start;
					Settings.Default.CurrentEnd = end;
				}
				else
				{
					Settings.Default.TimerRunning = false;
					Settings.Default.CurrentStart = DateTime.MinValue;
					Settings.Default.CurrentEnd = DateTime.MinValue;
				}

				Settings.Default.WindowHeight = RestoreBounds.Height;
				Settings.Default.WindowWidth = RestoreBounds.Width;
				Settings.Default.WindowLeft = RestoreBounds.Left;
				Settings.Default.WindowTop = RestoreBounds.Top;
				Settings.Default.WindowState = WindowState;
				Settings.Default.WindowSettingsSaved = true;
				Settings.Default.Save();
			}
			catch (Exception)
			{
			}
		}

		private void Window_Closed(object sender, EventArgs e)
		{
			if (updaterThread != null)
				updaterThread.Abort();
		}

		private void Data_Updated(object sender, EventArgs e)
		{
			if (updaterThread == null)
			{
				updaterThread = new Thread(new ThreadStart(UpdateInterface));
				updaterThread.Start();
			}
		}

		private void StartTimer()
		{
			if (updaterThread == null)
			{
				updaterThread = new Thread(new ThreadStart(UpdateInterface));
				updaterThread.Start();
			}
		}

		private void UpdateInterface()
		{
			while (true)
			{
				DispatcherOperation op = Dispatcher.BeginInvoke(new Action(delegate()
				{
					DateTime now = DateTime.Now;
					TimeSpan elapsed = now - start;
					TimeSpan remaining = end - now;
					TimeSpan total = end - start;

					if (remaining.Ticks <= 0)
					{
						if (!MainTextBox.IsFocused)
							MainTextBox.Text = "Timer expired";
						Title = "Orzeszek Timer";

						MainProgressBar.Value = 100;
						TaskbarUtility.SetProgressState(interopHelper.Handle, TaskbarProgressState.NoProgress);

						if (!notified)
							try
							{
								notified = true;
								SoundPlayer sp = new SoundPlayer(System.IO.Path.Combine("Sounds", Settings.Default.AlarmSound));
								sp.Play();
							}
							catch (Exception)
							{
							}

						if (Settings.Default.LoopTimer && Settings.Default.LastTimeSpan != TimeSpan.Zero)
						{
							start = DateTime.Now;
							end = start.Add(Settings.Default.LastTimeSpan);
							notified = false;
						}
						else
						{
							updaterThread.Abort();
							updaterThread = null;
						}
					}
					else
					{
						if (!MainTextBox.IsFocused)
							MainTextBox.Text = ToString(remaining);
						Title = MainTextBox.Text;

						MainProgressBar.Value = Math.Min(100.0, 100.0 * elapsed.Ticks / total.Ticks);
						TaskbarUtility.SetProgressValue(interopHelper.Handle, (ulong)MainProgressBar.Value, 100);
					}
				}
				));

				op.Wait();

				Thread.Sleep((int)Math.Max(Math.Min((end - start).TotalSeconds / MainProgressBar.ActualWidth, 1000), 10));
			}
		}

		private static object FromString(string s)
		{
			if (string.IsNullOrEmpty(s))
				return TimeSpan.Zero;

			DateTime dateTime;
			if (DateTime.TryParse(s, CultureInfo.CurrentCulture, DateTimeStyles.None, out dateTime))
				return dateTime;

			bool isDateTime = false;

			if (s.StartsWith("until ", StringComparison.CurrentCultureIgnoreCase))
			{
				s = s.Substring("until ".Length);
				isDateTime = true;
			}
			else if (s.StartsWith("to ", StringComparison.CurrentCultureIgnoreCase))
			{
				s = s.Substring("to ".Length);
				isDateTime = true;
			}

			if (isDateTime)
			{
				if (DateTime.TryParse(s, CultureInfo.CurrentCulture, DateTimeStyles.None, out dateTime))
					return dateTime;

				// @TODO: Parse other date formats.

				throw new FormatException();
			}

			double minutes;
			if (double.TryParse(s, out minutes))
				return TimeSpan.FromMinutes(minutes);

			try
			{
				TimeSpan ts = TimeSpan.Zero;
				string[] parts = Regex.Split(s, @"(?<=\d+\s*[a-zA-Z]+)\s*(?=\d+\s*[a-zA-Z]+)");
				foreach (string p in parts)
					if (!string.IsNullOrEmpty(p))
					{
						string[] subparts = Regex.Split(p, @"(?<=\d+)\s*(?=[a-zA-Z]+)");
						if (subparts.Length == 2)
							switch (subparts[1].ToLower()[0])
							{
								case 'd':
									ts = ts.Add(TimeSpan.FromDays(double.Parse(subparts[0])));
									break;
								case 'h':
									ts = ts.Add(TimeSpan.FromHours(double.Parse(subparts[0])));
									break;
								case 'm':
									ts = ts.Add(TimeSpan.FromMinutes(double.Parse(subparts[0])));
									break;
								case 's':
									ts = ts.Add(TimeSpan.FromSeconds(double.Parse(subparts[0])));
									break;
								default:
									throw new FormatException();
							}
						else
							throw new FormatException();
					}

				return ts;
			}
			catch (Exception)
			{
			}

			try
			{
				TimeSpan ts = TimeSpan.Zero;
				string[] parts = s.Split(new char[] { '.', ':', '-', ' ', '/', '\\', '\f', '\n', '\r', '\t', '\v', '\x85' }, StringSplitOptions.RemoveEmptyEntries);
				for (int i = 0; i < parts.Length; i++)
				{
					double d = double.Parse(parts[i]);
					switch (parts.Length - i)
					{
						case 1:
							ts = ts.Add(TimeSpan.FromSeconds(d));
							break;
						case 2:
							ts = ts.Add(TimeSpan.FromMinutes(d));
							break;
						case 3:
							ts = ts.Add(TimeSpan.FromHours(d));
							break;
						case 4:
							ts = ts.Add(TimeSpan.FromDays(d));
							break;
						default:
							throw new FormatException();
					}
				}

				return ts;
			}
			catch (Exception)
			{
			}

			TimeSpan timeSpan = TimeSpan.Zero;
			if (TimeSpan.TryParse(s, out timeSpan))
				return timeSpan;

			throw new FormatException();
		}

		private static string ToString(TimeSpan ts)
		{
			StringBuilder sb = new StringBuilder();

			if ((int)ts.TotalDays == 1)
				sb.AppendFormat("{0:D}\u00A0day ", (int)ts.TotalDays);
			else if ((int)ts.TotalDays != 0)
				sb.AppendFormat("{0:D}\u00A0days ", (int)ts.TotalDays);

			if ((int)ts.TotalHours == 1)
				sb.AppendFormat("{0:D}\u00A0hour ", ts.Hours);
			else if ((int)ts.TotalHours != 0)
				sb.AppendFormat("{0:D}\u00A0hours ", ts.Hours);

			if ((int)ts.TotalMinutes == 1)
				sb.AppendFormat("{0:D}\u00A0minute ", ts.Minutes);
			else if ((int)ts.TotalMinutes != 0)
				sb.AppendFormat("{0:D}\u00A0minutes ", ts.Minutes);

			if (ts.Seconds == 1)
				sb.AppendFormat("{0:D}\u00A0second", ts.Seconds);
			else
				sb.AppendFormat("{0:D}\u00A0seconds", ts.Seconds);

			return sb.ToString();
		}

		private void SoundMenuItem_Click(object sender, RoutedEventArgs e)
		{
			MenuItem soundMenuItem = sender as MenuItem;
			Settings.Default.AlarmSound = soundMenuItem.Tag == null ? string.Empty : (string)soundMenuItem.Tag;
			UpdateSoundMenuItems();
		}

		private void UpdateSoundMenuItems()
		{
			bool set = false;

			foreach (object o in MainContextMenu.Items)
			{
				MenuItem soundMenuItem = o as MenuItem;

				if (soundMenuItem != null)
					if (soundMenuItem.Tag != null && (string)soundMenuItem.Tag == Settings.Default.AlarmSound && !set)
					{
						soundMenuItem.IsChecked = true;
						set = true;
					}
					else if (soundMenuItem.Tag == null && string.IsNullOrEmpty(Settings.Default.AlarmSound) && !set)
					{
						soundMenuItem.IsChecked = true;
						set = true;
					}
					else
						soundMenuItem.IsChecked = false;
			}

			if (!set)
			{
				Settings.Default.AlarmSound = string.Empty;
				NoNotificationMenuItem.IsChecked = true;
			}
		}

		private void LoopTimerMenuItem_Click(object sender, RoutedEventArgs e)
		{
			Settings.Default.LoopTimer = LoopTimerMenuItem.IsChecked;
		}
	}
}